/*
 * pgm_fgets_realloc.c
 * Lettura immagine formato PGM con fgets.
 * Assunzioni:
 *  - lunghezza della riga illimitata
 *  - leggi_riga alloca la memoria del buffer caricato
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_LEVEL (256)

struct immagine {
  int magic_number;             // numero magico
  int max_gray_level;           // massimo livello di grigio
  int larg, alt;                // larghezza e altezza dell'immagine [pixel]
  int istogramma[MAX_LEVEL];    // vettore contenente l'istogramma
};

/* 
 * Attenzione che buf deve essere un puntatore, non buf[], in quanto
 * deve poter essere modificato.
 */
void analizza_riga(char *buf, struct immagine *pgm)
{
  char pixel[32];
  int n;

  while (sscanf(buf, "%31[^ ]%n", pixel, &n) == 1) {
    (pgm->istogramma[atoi(pixel)])++;
    buf += n;  /* aumenta ptr del numero di caratteri letti */
    if (*buf != ' ')
      break;   /* delimitatore non trovato; fine! */
    while (*buf == ' ' || *buf == '\n')
      ++buf;   /* salta il delimitatore e i successivi*/
  }
}

/**
 * FIO20-C. Avoid unintentional truncation when using fgets() or fgetws()
 * da https://www.securecoding.cert.org/confluence/pages/viewpage.action?pageId=3226
 */
char *leggi_riga(FILE *infile) {
  char temp[32];
  char *ret = malloc(sizeof(temp));
  char *end = ret;
  if (!ret) {
    return NULL;
  }
 
  while (fgets(temp, sizeof(temp), infile)) {
    size_t len = strlen(temp);
    const size_t full_size = end - ret + len;
    char *r_temp = realloc(ret, full_size + 1); /* NTBS */
    if (r_temp) {
      ret = r_temp;
      strcat(ret, temp);
      end = ret + full_size;
    } else {
      break;
    }
 
    if ((feof(infile)) || (temp[len - 1] == '\n')) {
      return ret;
    }
  }
  free(ret);
  return NULL;
}

int pgm_istogramma(struct immagine *pgm, FILE *infile)
{
  int n, i;
  char *buf = NULL;

  for (i = 0; i < MAX_LEVEL; i++)
    pgm->istogramma[i] = 0;

  /* lettura e parsing della riga contenente il magic number */
  buf = leggi_riga(infile);
  n = sscanf(buf, "P%i", &pgm->magic_number);
  if ((n != 1) || (pgm->magic_number != 2)) return 1;
  free(buf);

  /* lettura e parsing della riga contenente le dimensioni */
  buf = leggi_riga(infile);
  if (sscanf(buf, "%i %i", &pgm->larg, &pgm->alt) != 2) return 1;
  free(buf);

  /* lettura e parsing della riga contenente i livelli di grigio */
  buf = leggi_riga(infile);
  if (sscanf(buf, "%i", &pgm->max_gray_level) != 1) return 1;
  free(buf);

  /* lettura e parsing del resto del file */
  while ((buf = leggi_riga(infile))) {
    analizza_riga(buf, pgm);
    free(buf);
  }

  return 0;
}

void print_istogramma(struct immagine *pgm, FILE *outfile)
{
  int livello;

  for (livello = 0; livello < pgm->max_gray_level; livello++)
    fprintf(outfile, "%3i: %i\n", livello, pgm->istogramma[livello]);
}

int main(int argc, char *argv[])
{
  struct immagine pgm = {.alt = 0, .larg = 0, .max_gray_level = 0};
  FILE *infile;

  if (argc != 2) {
    printf("uso: %s <nome_file_pgm>\n", argv[0]);
    return 1;
  }
  if (!(infile = fopen(argv[1], "r"))) {
    printf("Errore nell'apertura di %s\n", argv[1]);
    return 1;
  }
  if (pgm_istogramma(&pgm, infile) != 0) {
    puts("Errore nella lettura del file.");
    return 1;
  }

  print_istogramma(&pgm, stdout);

  return 0;
}
